/*******************************************************************************

  Copyright(c) 2003-2004 Aurelien Reynaud.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The full GNU General Public License is included in this distribution in the
  file called COPYING.

*******************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "ggiterm.h"
#include "debug.h"


/* Global variables */
int (*backend_write_char) (int, int, cell_t, int) = NULL;
wchar_t (*backend_get_charmap) () = NULL;

/* External functions */
int native_init (int*, int*);
int native_write_char (int, int, cell_t, int);
wchar_t native_get_charmap ();
void native_exit ();


#ifdef HAVE_FREETYPE

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

/* External functions */
int freetype_init (char *, int, int*, int*);
int freetype_write_char (int, int, cell_t, int);
wchar_t freetype_get_charmap ();
void freetype_exit ();

#endif /* HAVE_FREETYPE */

/*
 * Determines and initializes the best possible backend
 * Fills in cell_h, cell_w and backend_write_char
 * Returns 0 on success, 1 on failure
 */
int backend_init (char *font_file, int fontsize, int *cell_w, int *cell_h)
{
	int err = 1;

	debug (DEBUG_FUNCTION,
	       "called with args (font_file=%s, fontsize=%d, *cell_w, *cell_h)",
	       font_file, fontsize);
	if (font_file != NULL) {
#ifdef HAVE_FREETYPE
		debug (DEBUG_INIT, "Trying Freetype backend");
		err = freetype_init (font_file, fontsize, cell_w, cell_h);
		if (!err) {
			debug (DEBUG_INIT, "Using Freetype backend");
			backend_write_char = &freetype_write_char;
			backend_get_charmap = &freetype_get_charmap;
		}
#else /* HAVE_FREETYPE */
		error ("Freetype support not compiled in");
#endif /* HAVE_FREETYPE */
		if (err) {
			debug (DEBUG_INIT, "Cannot use Freetype backend");
			err = native_init (cell_w, cell_h);
			if (!err) {
				error ("Falling back to native GGI "
		 	 	       "backend (ASCII only)");
				backend_write_char = &native_write_char;
				backend_get_charmap = &native_get_charmap;
			}
		}
	} else {
		debug (DEBUG_INIT, "No font file provided");
		err = native_init (cell_w, cell_h);
		if (!err) {
			debug (DEBUG_INIT,
			       "Using native GGI backend (ASCII only)");
			backend_write_char = &native_write_char;
			backend_get_charmap = &native_get_charmap;
		}
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return err;
}

/*
 * Frees the resources associated with the backend in use
 */
void backend_exit ()
{
	debug (DEBUG_FUNCTION, "called with no args");
#ifdef HAVE_FREETYPE
	if (backend_write_char == &freetype_write_char) {
		freetype_exit ();
	}
#endif /* HAVE_FREETYPE */
	if (backend_write_char == &native_write_char) {
		native_exit ();
	}
	debug (DEBUG_FUNCTION, "Leaving");
}
