/*******************************************************************************

  Copyright(c) 2003-2004 Aurelien Reynaud.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The full GNU General Public License is included in this distribution in the
  file called COPYING.

*******************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "ggiterm.h"
#include "debug.h"
#include "terminal.h"


#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

#include <ggi/ggi.h>

/* Global variables */
FT_Library library;
FT_Face face;
FT_Int32 my_render_flags;
int underline_y, underline_h;
ggi_color *colortab = NULL;
void *mypixelbuf = NULL;
int cursor_x0, cursor_y0;

/* External functions */
wchar_t freetype_get_altchar (wchar_t);

extern ggi_visual_t vis;
extern ggi_color palette[];
extern ggi_pixel color[];

int ft_width, ft_height;


int freetype_validate_font ()
{
	debug (DEBUG_FUNCTION, "called with no args");
	
	debug (DEBUG_INIT, "Font family: %s", face->family_name);
	if (FT_IS_SCALABLE (face)) {
		debug (DEBUG_INIT, "font is scalable");
	} else {
		debug (DEBUG_INIT,    "font is not scalable");
	}
	if (FT_IS_FIXED_WIDTH (face)) {
		debug (DEBUG_INIT,
		       "font is fixed width");
	} else {
		debug (DEBUG_INIT,
		       "font is not fixed width");
		error ("font is not fixed width");
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	debug (DEBUG_INIT, "%ld glyphs in font",
	       face->num_glyphs);
	debug (DEBUG_INIT,
	       "global metrics in font units:");
	debug (DEBUG_INIT,
	       "BBox width  = %ld to %ld (%ld)",
	       face->bbox.xMin, face->bbox.xMax,
	       face->bbox.xMax - face->bbox.xMin);
	debug (DEBUG_INIT,
	       "BBox height = %ld to %ld (%ld)",
	       face->bbox.yMin, face->bbox.yMax,
	       face->bbox.yMax - face->bbox.yMin);
	debug (DEBUG_INIT, "text height = %d",
	       face->height);
	debug (DEBUG_INIT,
	       "maximum advance width = %d",
	       face->max_advance_width);
	debug (DEBUG_INIT, "units per EM = %d",
	       face->units_per_EM);
	debug (DEBUG_INIT, "ascender = %d",
	       face->ascender);
	debug (DEBUG_INIT, "underline position = %d",
	       face->underline_position);
	debug (DEBUG_INIT, "underline thickness = %d",
	       face->underline_thickness);
	if (!face->underline_thickness) {
		error ("WARNING! font does not allow underlining. "
		       "Going on anyway...");
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return 0;
}

int freetype_write_char (int x, int y, cell_t charcode, int blinking)
{
	int i, j, colorlen, err;
	unsigned char *tab;
	unsigned char val;
	ggi_color bg, fg;
	int cur_bgcolor, cur_fgcolor, dummy_color;

	debug (DEBUG_FUNCTION,
	       "called with args (x=%d, y=%d, charcode=[cell_t], blinking=%d",
	       x, y, blinking);
	
	cur_bgcolor = charcode.bgcolor;
	cur_fgcolor = charcode.fgcolor;
	if (charcode.mode & MODE_ALTCHARSET) {
		charcode.charcode = freetype_get_altchar (charcode.charcode);
	}
	if (charcode.mode & MODE_REVERSE) {
		dummy_color = cur_bgcolor;
		cur_bgcolor = cur_fgcolor;
		cur_fgcolor = dummy_color;
	}
	if (charcode.mode & MODE_BOLD) {
		cur_fgcolor += 8;
	}
	if (charcode.mode & MODE_INVISIBLE) {
		cur_fgcolor = cur_bgcolor;
	}
	if ((charcode.mode & MODE_BLINK) && blinking) {
		dummy_color = cur_bgcolor;
		cur_bgcolor = cur_fgcolor;
		cur_fgcolor = dummy_color;
	}
	
	err = FT_Load_Char (face, charcode.charcode, my_render_flags);
	if (err) {
		error ("No charmap defined or unknown character code");
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	debug (DEBUG_FT, "Glyph rendered to bitmap");

	debug (DEBUG_FT, "Bitmap fields:");
	debug (DEBUG_FT, "%d rows %d pixels wide",
	       face->glyph->bitmap.rows, face->glyph->bitmap.width);
	debug (DEBUG_FT, "pitch = %d",
	       face->glyph->bitmap.pitch);
	debug (DEBUG_FT, "Number of grays: %d",
	       face->glyph->bitmap.num_grays);

	debug (DEBUG_FT, "Bitmap values:");
	tab = face->glyph->bitmap.buffer;
	colorlen = 0;
	bg = palette[cur_bgcolor];
	fg = palette[cur_fgcolor];
	for (j = 0; j < face->glyph->bitmap.rows; j++) {
		debug (DEBUG_FT, "%3d: ", j);
		for (i = 0; i < face->glyph->bitmap.width; i++) {
			if (my_render_flags & FT_LOAD_MONOCHROME)
				/* Maybe replace 128>> with 1<< on bigendian */
				val = (tab[i/8] & (128>>(i%8)) ? 255 : 0);
			else val = tab[i];
			debug (DEBUG_FT, " %3d", val);
			colortab[colorlen].r = bg.r + ((fg.r-bg.r) * val) / 255;
			colortab[colorlen].g = bg.g + ((fg.g-bg.g) * val) / 255;
			colortab[colorlen].b = bg.b + ((fg.b-bg.b) * val) / 255;
			colorlen++;
		}
		debug (DEBUG_FT, "\n");
		tab += face->glyph->bitmap.pitch;
	}

	/* Erase background */
	ggiSetGCForeground (vis, color[cur_bgcolor]);
	ggiDrawBox (vis, x, y, ft_width, ft_height);
	
	err = ggiPackColors (vis, mypixelbuf, colortab, colorlen);
	if (err) {
		error ("PackColors failed");
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	debug (DEBUG_FT, "PackColors succeeded");

	err = ggiPutBox (vis,
			x + cursor_x0 + face->glyph->bitmap_left,
			y + cursor_y0 - face->glyph->bitmap_top,
			face->glyph->bitmap.width,
			face->glyph->bitmap.rows,
			mypixelbuf);
	if (err) {
		error ("PutBox failed");
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	debug (DEBUG_FT, "PutBox succeeded");

	if ((charcode.mode & MODE_UNDERLINE) && underline_h != 0) {
		ggiSetGCForeground (vis, color[cur_fgcolor]);
		ggiDrawBox (vis,
		            x, y + cursor_y0 + underline_h,
		            ft_width, underline_h);
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return 0;
}

int freetype_init (char *font_file, int fontsize, int *cell_w, int *cell_h)
{
	int err;
	ggi_mode current_mode;
	long memory_size;
	int mypixelsize;
	const ggi_pixelformat *mypixelformat;

	debug (DEBUG_FUNCTION,
	       "called with args (font_file=%s, fontsize=%d, *cell_w, *cell_h",
	       font_file, fontsize);
	
	err = FT_Init_FreeType (&library);
	if (err) {
		error ("an error occurred during library initialization ...");
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	debug (DEBUG_INIT, "library initialized");

	err = FT_New_Face (library, font_file, 0, &face);
	if (err == FT_Err_Unknown_File_Format) {
		error ("Font file %s could be opened and read, but it "
		       "appears that its font format is unsupported ...",
		       font_file);
		FT_Done_FreeType (library);
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	} else if (err) {
		error ("Font file %s could not be opened or read, or "
		       "simply is broken",
		       font_file);
		FT_Done_FreeType (library);
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;

	}
	debug (DEBUG_INIT,
	       "Font file %s successfully opened",
	       font_file);
	err = freetype_validate_font ();
	if (err) {
		error ("Font not suitable!");
		FT_Done_FreeType (library);
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	err = FT_Set_Pixel_Sizes (face, fontsize, 0);
	if (err) {
		error ("Couldn't set pixel size");
		FT_Done_FreeType (library);
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	
	/* Do we do aliasing? */
	ggiGetMode (vis, &current_mode);
	if (GT_DEPTH(current_mode.graphtype) > 8) {
		debug (DEBUG_INIT,
	               "Color depth is %d: Will do aliasing",
		       GT_DEPTH(current_mode.graphtype));
		my_render_flags = FT_LOAD_RENDER;
	} else {
		debug (DEBUG_INIT,
		       "Color depth is %d: Won't do aliasing",
		       GT_DEPTH(current_mode.graphtype));
		my_render_flags = FT_LOAD_RENDER | FT_LOAD_MONOCHROME;
	}
	
	/*ft_width = ceil (fontsize * (float)(face->bbox.xMax -
			     face->bbox.xMin) / face->units_per_EM);*/
	/*ft_width = ceil (fontsize * (float)(face->max_advance_width) / face->units_per_EM);*/
	ft_width = (fontsize * face->max_advance_width) % face->units_per_EM ?
	           (fontsize * face->max_advance_width) / face->units_per_EM +1 :
	           (fontsize * face->max_advance_width) / face->units_per_EM;
	ft_height = (fontsize * face->height) % face->units_per_EM ?
	            (fontsize * face->height) / face->units_per_EM +1 :
	            (fontsize * face->height) / face->units_per_EM;
	/*ft_height = ceil (fontsize * (float)(face->bbox.yMax -
			     face->bbox.yMin) / face->units_per_EM);*/
	debug (DEBUG_INIT,
	       "Cell size successfully set to %d x %d", ft_width, ft_height);
	cursor_x0 = 0;
	cursor_y0 = (face->bbox.yMax * fontsize) % face->units_per_EM ?
	            (face->bbox.yMax * fontsize) / face->units_per_EM + 1 :
	            (face->bbox.yMax * fontsize) / face->units_per_EM;

	underline_y = fontsize *
	              (face->underline_position - (face->underline_thickness / 2)) /
		      face->units_per_EM;
	underline_h = (fontsize * face->underline_thickness) % face->units_per_EM ?
	              (fontsize * face->underline_thickness) / face->units_per_EM + 1 :
		      (fontsize * face->underline_thickness) / face->units_per_EM;
	debug (DEBUG_INIT,
	       "Underline starts at %d pixels from baseline", underline_y);
	debug (DEBUG_INIT,
	       "Underline is %d pixels thick", underline_h);
	
	/* freetype-specific memory allocation */
	mypixelformat = ggiGetPixelFormat (vis);
	mypixelsize = mypixelformat->size;
	memory_size = ft_height * ft_width * sizeof (ggi_color);
	colortab = malloc (memory_size);
	memory_size += (mypixelsize * ft_height * ft_width + 7)/8;
	mypixelbuf = malloc ((mypixelsize * ft_height * ft_width + 7)/8);
	if (mypixelbuf == NULL) {
		error ("Couldn't allocate memory");
		FT_Done_FreeType (library);
		debug (DEBUG_FUNCTION, "Leaving");
		return 1;
	}
	debug (DEBUG_INIT,
	       "%ld bytes allocated for freetype rendering engine",
	       memory_size);
	*cell_w = ft_width;
	*cell_h = ft_height;
	
	debug (DEBUG_FUNCTION, "Leaving");
	return 0;
}

wchar_t freetype_get_charmap ()
{
	static FT_UInt gindex;
	static wchar_t last = 0;
	
	debug (DEBUG_FUNCTION, "called with no args");
	if (last == 0) {
		last = FT_Get_First_Char (face, &gindex);
	} else {
		if (gindex != 0) {
			last = FT_Get_Next_Char (face, last, &gindex);
		} else {
			last = 0;
		}
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return last;
}

void freetype_exit ()
{
	debug (DEBUG_FUNCTION, "called with no args");
	free (colortab);
	free (mypixelbuf);
	FT_Done_FreeType (library);
	debug (DEBUG_FUNCTION, "Leaving");
}
