/*******************************************************************************

  Copyright(c) 2003-2004 Aurelien Reynaud.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The full GNU General Public License is included in this distribution in the
  file called COPYING.

*******************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "ggiterm.h"
#include "debug.h"
#include "terminal.h"

#include <string.h>


/*
 * Parses the sequence beginning at buffer and processes its contents
 * to fill the provided array of parameters
 * Return values:
 * 	0: a valid (possibly empty) sequence has been processed
 * 	1: the sequence is valid but contains out of range values
 * 	2: uncomplete sequence, retry with more data
 * If success, the buffer is increased by the number of bytes processed and
 * the buffer size is decreased accordingly. Else they are left untouched.
 */
int parse_ansi_params (unsigned char **buffer, size_t *buffer_size,
                         unsigned int **p, int *nb_param)
{
	static unsigned int param[MAX_PARAM_NUM];
	size_t buf_size;
	unsigned char *buf;
	int i, err=0;

	debug (DEBUG_FUNCTION, "Entering");
	buf = *buffer;
	buf_size = *buffer_size;
	
	for (i=0; i<MAX_PARAM_NUM; i++) param[i] = 0;
	i = 0;
		
	while (buf_size && !err) {
		switch (buf[0]) {
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				/* A little bit complicated but necessary to
			           avoid integer overflow */
				if (param[i]<= MAX_PARAM_VALUE/10) {
					param[i] *= 10;
				} else err = 1;
				if (param[i] <= MAX_PARAM_VALUE - (buf[0]-48)) {
					param[i] += (buf[0]-48);
				} else err = 1;
				break;
			case ';':
				i++;
				if (i >= MAX_PARAM_NUM) err = 1;
				break;
			default:
				goto stop;
		}
		buf++; buf_size--;
	}
	
stop:
	if (!buf_size) err = 2;
	if (!err) {
		if (buf_size == *buffer_size) *nb_param = 0;
		else *nb_param = i+1;
		*buffer = buf;
		*buffer_size = buf_size;
		*p = param;
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return err;
}

/*
 * Executes the sequence beginning at the given buffer.
 * The first character in the buffer immediately follows the \E[?
 * Return values:
 * 	0: a valid (possibly empty) sequence has been processed
 * 	1: the sequence is valid but contains out of range values
 * 	2: uncomplete sequence, retry with more data
 * If success, the buffer is increased by the number of bytes processed and
 * the buffer size is decreased accordingly. Else they are left untouched.
 */
int parse_ansi_qmark (unsigned char **buffer, size_t *buffer_size)
{
	size_t buf_size;
	unsigned char *buf;
	unsigned int *param;
	int i, err=0, nb_param;

	debug (DEBUG_FUNCTION, "Entering");
	buf = *buffer;
	buf_size = *buffer_size;
	
	err = parse_ansi_params (&buf, &buf_size, &param, &nb_param);
	if (err) {
		debug (DEBUG_FUNCTION, "Leaving");
		return err;
	}
	
	switch (buf[0]) {
		case 'h':
			for (i = 0; i < nb_param; i++) {
				switch (param[i]) {
					case 1:
						ignore_keypad_num = 1;
						break;
					case 7:
						cursor_set_autowrap ();
						break;
					/* other modes are not implemented */
					case 3:
					case 4:
					case 5:
					case 6:
					case 8:
					case 9:
						break;
					case 25:	/* ggiterm specific */
						debug (DEBUG_OUTPUT, "cnorm (cursor visible again)");
						cursor_unset_mode (CURSOR_INVISIBLE);
						break;
					default:
						err = 1;
				}
			}
			break;
		case 'l':
			for (i = 0; i < nb_param; i++) {
				switch (param[i]) {
					case 1:
						ignore_keypad_num = 0;
						break;
					case 7:
						cursor_unset_autowrap ();
						break;
					/* other modes are not implemented */
					case 2:
					case 3:
					case 4:
					case 5:
					case 6:
					case 8:
					case 9:
						break;
					case 25:	/* ggiterm specific */
						debug (DEBUG_OUTPUT, "civis (make cursor invisible)");
						cursor_set_mode (CURSOR_HIDE | CURSOR_INVISIBLE);
						break;
					default:
						err = 1;
				}
			}
			break;
	}
	buf++; buf_size--;
		
	if (!err) {
		*buffer = buf;
		*buffer_size = buf_size;
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return err;
}

/*
 * Executes the ANSI sequence beginning at the given buffer.
 * The first character in the buffer immediately follows the \E[
 * Return values:
 * 	0: a valid special sequence has been found and executed
 * 	1: not a known special sequence
 * 	2: uncomplete sequence, retry with more data
 * 	3: can't execute because internal buffer is full, read() it and retry
 * If success, the buffer is increased by the number of bytes processed and
 * the buffer size is decreased accordingly. Else they are left untouched.
 */
int parse_ansi_sequences (unsigned char **buffer, size_t *buffer_size)
{
	size_t buf_size;
	unsigned char *buf;
	unsigned int *param;
	int i, err, nb_param;
	int line, column;
	char answer[BUFFER_SIZE];

	debug (DEBUG_FUNCTION, "Entering");
	buf = *buffer;
	buf_size = *buffer_size;
	
	if (buf[0] == '?') {
		buf++; buf_size--;
		err = parse_ansi_qmark (&buf, &buf_size);
		if (!err) {
			*buffer = buf;
			*buffer_size = buf_size;
		}
		debug (DEBUG_FUNCTION, "Leaving");
		return err;
	}
	
	err = parse_ansi_params (&buf, &buf_size, &param, &nb_param);
	if (err) {
		debug (DEBUG_FUNCTION, "Leaving");
		return err;
	}
	
	switch (buf[0]) {
		case '@':
			if (nb_param) {
				debug (DEBUG_OUTPUT, "ich %d (insert characters)", param[0]);
				cell_insert (param[0]);
			} else {
				debug (DEBUG_OUTPUT, "ich 1 (insert characters)");
				cell_insert (1);
			}
			break;
		case 'A':
			cursor_get_position (&column, &line);
			if (nb_param) {
				debug (DEBUG_OUTPUT, "cuu %d (move cursor up)", param[0]);
				cursor_move (column, line - param[0]);
			} else {
				debug (DEBUG_OUTPUT, "cuu1 (move cursor up)");
				cursor_move (column, line - 1);
			}
			break;
		case 'B':
			cursor_get_position (&column, &line);
			if (nb_param) {
				debug (DEBUG_OUTPUT, "cud %d (move cursor down)", param[0]);
				cursor_move (column, line + param[0]);
			} else {
				debug (DEBUG_OUTPUT, "cud1 (move cursor down)");
				cursor_move (column, line + 1);
			}
			break;
		case 'C':
			cursor_get_position (&column, &line);
			if (nb_param) {
				debug (DEBUG_OUTPUT, "cuf %d (move cursor right)", param[0]);
				cursor_move (column + param[0], line);
			} else {
				debug (DEBUG_OUTPUT, "cuf1 (move cursor right)");
				cursor_move (column + 1, line);
			}
			break;
		case 'D':
			cursor_get_position (&column, &line);
			if (nb_param) {
				debug (DEBUG_OUTPUT, "cub %d (move cursor left)", param[0]);
				cursor_move (column - param[0], line);
			} else {
				debug (DEBUG_OUTPUT, "cub1 (move cursor left)");
				cursor_move (column - 1, line);
			}
			break;
		case 'F':
			if (nb_param) {
				debug (DEBUG_OUTPUT, "rep %d (repeat last character)", param[0]);
				text_repeat_char (param[0]);
			} else {
				debug (DEBUG_OUTPUT, "rep 1 (repeat last character)");
				text_repeat_char (1);
			}
			break;
		case 'H':
		case 'f':
			switch (nb_param) {
				case 0:
					debug (DEBUG_OUTPUT, "home/cup 1 1 (move cursor to)");
					cursor_move (0, 0);
					break;
				case 1:
					debug (DEBUG_OUTPUT, "cup %d 1 (move cursor to)", param[0]);
					cursor_move (0, param[0] - 1);
					break;
				default:
					debug (DEBUG_OUTPUT, "cup %d %d (move cursor to)", param[0], param[1]);
					cursor_move (param[1] - 1, param[0] - 1);
			}
			break;
		case 'J':
			if (!nb_param) {
				debug (DEBUG_OUTPUT, "ed (erase to end of display)");
				screen_erase_from ();
				break;
			}
			if (nb_param == 1) {
				switch (param[0]) {
					case 0:
						debug (DEBUG_OUTPUT, "ed (erase to end of display)");
						screen_erase_from ();
						break;
					case 1:
						debug (DEBUG_OUTPUT, "vt100 (erase from beginning of display)");
						screen_erase_to ();
						break;
					case 2:
						debug (DEBUG_OUTPUT, "vt100 (erase display)");
						screen_erase ();
						break;
					default:
						err = 1;
				}
			} else err = 1;
			break;
		case 'K':
			if (!nb_param) {
				debug (DEBUG_OUTPUT, "el (erase to end of line)");
				line_erase_from ();
				break;
			}
			if (nb_param == 1) {
				switch (param[0]) {
					case 0:
						debug (DEBUG_OUTPUT, "el (erase to end of line)");
						line_erase_from ();
						break;
					case 1:
						debug (DEBUG_OUTPUT, "el1 (erase from beginning of line)");
						line_erase_to ();
						break;
					case 2:
						debug (DEBUG_OUTPUT, "vt100 (erase line)");
						line_erase ();
						break;
					default:
						err = 1;
				}
			} else err = 1;
			break;
		case 'P':
			if (nb_param) {
				debug (DEBUG_OUTPUT, "dch %d (delete characters)", param[0]);
				cell_delete (param[0]);
			} else {
				debug (DEBUG_OUTPUT, "dch 1 (delete characters)");
				cell_delete (1);
			}
			break;
		case 'S':
			if (nb_param) {
				debug (DEBUG_OUTPUT, "indn %d (scroll forward n lines", param[0]);
				scroll_up (param[0]);
			} else {
				debug (DEBUG_OUTPUT, "indn 1 (scroll forward 1 line");
				scroll_up (1);
			}
			break;
		case 'T':
			if (nb_param) {
				debug (DEBUG_OUTPUT, "rin %d (scroll back n lines", param[0]);
				scroll_down (param[0]);
			} else {
				debug (DEBUG_OUTPUT, "rin 1 (scroll back 1 line");
				scroll_down (1);
			}
			break;
		case 'X':
			if (nb_param) {
				debug (DEBUG_OUTPUT, "ech %d (erase characters)", param[0]);
				cell_erase (param[0]);
			} else {
				debug (DEBUG_OUTPUT, "ech 1 (erase characters)");
				cell_erase (1);
			}
			break;
		case 'Z':
			debug (DEBUG_OUTPUT, "cbt (go to previous tab stop)");
			tabstop_prev ();
			break;
		case 'c':
			if (nb_param == 1 && param[0] == 0) {
				debug (DEBUG_OUTPUT, "\\005/u9/identify");
				/* Enquire sequence sent to identify the terminal type
				   Should answer by sending the u8 user string	*/
				if (buffer_send_string ("\033[?1;0c")) err = 3;
			}
			else err = 1;
			break;
		case 'g':
			if (nb_param == 1) {
				switch (param[0]) {
					case 3:
						debug (DEBUG_OUTPUT, "tbc (clear all tab stops)");
						tabstop_clear_all ();
						break;
					case 0:
						debug (DEBUG_OUTPUT, "vt100 (clear current tab stop)");
						tabstop_clear ();
						break;
					default:
						err = 1;
				}
			} else err = 1;
			break;
		case 'h':
			if (nb_param != 1 || param[0] != 20) err = 1;
			break;
		case 'l':
			if (nb_param != 1 || param[0] != 20) err = 1;
			break;
		case 'm':
			if (!nb_param) {
				debug (DEBUG_OUTPUT, "sgr/rmso (all modes off except altcharset)");
				text_unset_mode (MODE_BLINK|MODE_BOLD|MODE_INVISIBLE|MODE_REVERSE|MODE_UNDERLINE|MODE_ALTCHARSET);
				text_set_bgcolor (9);
				text_set_fgcolor (9);
			}
			for (i = 0; i < nb_param; i++) {
				switch (param[i]) {
					case 0:
						debug (DEBUG_OUTPUT, "sgr/rmso (all modes off except altcharset)");
						text_unset_mode (MODE_BLINK|MODE_BOLD|MODE_INVISIBLE|MODE_REVERSE|MODE_UNDERLINE|MODE_ALTCHARSET);
						text_set_bgcolor (9);
						text_set_fgcolor (9);
						break;
					case 1:
						debug (DEBUG_OUTPUT, "bold/sgr (bold mode on)");
						text_set_mode (MODE_BOLD);
						break;
					case 4:
						debug (DEBUG_OUTPUT, "smul/sgr (underline mode on)");
						text_set_mode (MODE_UNDERLINE);
						break;
					case 5:
						debug (DEBUG_OUTPUT, "blink/sgr (blink mode on)");
						text_set_mode (MODE_BLINK);
						break;
					case 7:
						debug (DEBUG_OUTPUT, "rev/smso/sgr (reverse video mode on)");
						text_set_mode (MODE_REVERSE);
						break;
					case 8:
						debug (DEBUG_OUTPUT, "invis/sgr (invisible mode on)");
						text_set_mode (MODE_INVISIBLE);
						break;
					default:
						if (param[i]/10 == 3) {
							debug (DEBUG_OUTPUT, "setab %d (set foreground color)", param[i]-30);
							text_set_fgcolor (param[i]-30);
							break;
						}
						if (param[i]/10 == 4) {
							debug (DEBUG_OUTPUT, "setab %d (set background color)", param[i]-40);
							text_set_bgcolor (param[i]-40);
							break;
						}
						err = 1;
				}
			}
			break;
		case 'n':
			if (nb_param == 1) {
				switch (param[0]) {
					case 5: /* Request terminal status
						   \E[0n if OK, \E[3n if NOK
					          (but we're always OK, aren't we?) */
						debug (DEBUG_OUTPUT, "status request");
						if (buffer_send_string ("\033[0n")) err = 3;
						break;
					case 6: /* Cursor position request */
						debug (DEBUG_OUTPUT, "cursor position request");
						cursor_get_position (&column, &line);
						sprintf (answer, "\033[%d;%dR", line + 1, column + 1);
						if (buffer_send_string (answer)) err = 3;
						break;
					default:
						err = 1;
				}
			} else err = 1;
			break;
		case 'r':
			switch (nb_param) {
				case 0:
					change_region (-1, -1);
					break;
				case 1:
					change_region (param[0]-1, -1);
					break;
				default:
					change_region (param[0]-1, param[1]);
			}
			break;
		default:
			err = 1;
	}
	buf++; buf_size--;
	
	if (!err) {
		*buffer = buf;
		*buffer_size = buf_size;
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return err;
}

/*
 * Executes the escape sequence at the given buffer.
 * The first character in the buffer immediately follows the ESC character
 * Return values:
 * 	0: a valid special sequence has been found and executed
 * 	1: not a known special sequence
 * 	2: uncomplete sequence, retry with more data
 * 	3: can't execute because internal buffer is full, read() it and retry
 * If success, the buffer is increased by the number of bytes processed and
 * the buffer size is decreased accordingly. Else they are left untouched.
 */
int parse_escape_sequences (unsigned char **buffer, size_t *buffer_size)
{
	size_t buf_size;
	unsigned char *buf;
	int err=0;
	int line, column;

	debug (DEBUG_FUNCTION, "Entering");
	if (*buffer_size == 0) {
		debug (DEBUG_FUNCTION, "Leaving");
		return 2;
	}
	buf = *buffer;
	buf_size = *buffer_size;
	
	buf++; buf_size--;
	switch ((*buffer)[0]) {
		case '[':
			err = parse_ansi_sequences (&buf, &buf_size);
			break;
		case '(': /* Character set designator G0 */
			if (buf_size < 1) {
				err = 2;
				break;
			}
			buf ++; buf_size --;
			switch ((*buffer)[1]) {
				case '0': /* line drawing set */
					text_set_mode (MODE_ALTCHARSET);
					break;
				case '1': /* alternate character ROM */
				case '2': /* alternate character ROM (gfx) */
				case 'A': /* United Kingdom */
				case 'B': /* US ASCII */
					text_unset_mode (MODE_ALTCHARSET);
					break;
				default:
					err = 1;
			}
			break;
		case ')': /* Character set designator G1 */
			if (buf_size < 1) {
				err = 2;
				break;
			}
			buf ++; buf_size --;
			switch ((*buffer)[1]) {
				case '0': /* line drawing set */
				case '1': /* alternate character ROM */
				case '2': /* alternate character ROM (gfx) */
				case 'A': /* United Kingdom */
				case 'B': /* US ASCII */
					break;
				default:
					err = 1;
			}
			break;
		case 'D':
			debug (DEBUG_OUTPUT, "vt100 (scroll forward 1 line)");
			scroll_up (1);
			break;
		case 'E':
			debug (DEBUG_OUTPUT, "vt100 (CR+LF)");
			cursor_get_position (&column, &line);
			cursor_move (0, line + 1);
			break;
		case 'H':
			debug (DEBUG_OUTPUT, "hts (set tab stop)");
			tabstop_set ();
			break;
		case 'M':
			debug (DEBUG_OUTPUT, "vt100 (scroll back 1 line)");
			scroll_down (1);
			break;
		case 'Z': /* identify, same as \E[c */
			debug (DEBUG_OUTPUT, "\\005/u9/identify");
			/* Enquire sequence sent to identify the terminal type
			   Should answer by sending the u8 user string	*/
			if (buffer_send_string ("\033[?1;0c")) err = 3;
			break;
		case '>':
			/*force_cursor_mode = 1;*/
			break;
		case '=':
			/*force_cursor_mode = 0;*/
			break;
		case '7':
			save_state ();
			break;
		case '8':
			restore_state ();
			break;
		default:
			err = 1;
	}

	if (!err) {
		*buffer = buf;
		*buffer_size = buf_size;
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return err;
}

/*
 * Executes the special sequence beginning at the given buffer.
 * Return values:
 * 	0: a valid special sequence has been found and executed
 * 	1: not a known special sequence
 * 	2: uncomplete sequence, retry with more data
 * 	3: can't execute because internal buffer is full, read() it and retry
 * If success, the buffer is increased by the number of bytes processed and
 * the buffer size is decreased accordingly. Else they are left untouched.
 */
int parse_special_chars (unsigned char **buffer, size_t *buffer_size)
{
	size_t buf_size;
	unsigned char *buf;
	int err=0;
	int line, column;

	debug (DEBUG_FUNCTION, "Entering");
	buf = *buffer;
	buf_size = *buffer_size;
	buf++; buf_size--;
	
	switch ((*buffer)[0]) {
		case 000: /* ignored */
			break;
		case 005:
			debug (DEBUG_OUTPUT, "\\005/u9/identify");
			/* Enquire sequence sent to identify the terminal type
			   Should answer by sending the u8 user string	*/
			if (buffer_send_string ("\033[?1;0c")) err = 3;
			break;
		case 007:		/* Bell (7) */
			debug (DEBUG_OUTPUT, "\\007 (bell)");
			screen_flash ();
			break;
		case 010:		/* Backspace */
			debug (DEBUG_OUTPUT, "\\010 (backspace)");
			cursor_get_position (&column, &line);
			cursor_move (column - 1, line);
			break;
		case 011:		/* Tab expansion */
			debug (DEBUG_OUTPUT, "\\011/ht (go to next tab stop)");
			tabstop_next ();
			break;
		case 012:		/* Line feed (10) */
		case 013:
		case 014:
			debug (DEBUG_OUTPUT, "\\012/\\013/\\014 (line feed)");
			cursor_get_position (&column, &line);
			cursor_move (column, line + 1);
			break;
		case 015:		/* Carriage return (13) */
			debug (DEBUG_OUTPUT, "\\015 (carriage return)");
			cursor_get_position (&column, &line);
			cursor_move (0, line);
			break;
		case 016:
			debug (DEBUG_OUTPUT, "smacs/sgr (altcharset mode on)");
			text_set_mode (MODE_ALTCHARSET);
			break;
		case 017:
			debug (DEBUG_OUTPUT, "rmacs/sgr/sgr0 (altcharset mode off)");
			text_unset_mode (MODE_ALTCHARSET);
			break;
		case 033:		/*  Escape character */
			err = parse_escape_sequences (&buf, &buf_size);
			break;
		case 0177:		/* for FreeBSD */
			debug (DEBUG_OUTPUT, "\\0177 (FreeBSD backspace)");
			cursor_get_position (&column, &line);
			cursor_move (column - 1, line);
			break;
		default:
			err = 1;
	}
	if (!err) {
		debug (DEBUG_OUTPUT, "Executed special sequence was: %s",
		       binary_to_printable (*buffer, *buffer_size-buf_size));
		*buffer = buf;
		*buffer_size = buf_size;
	}
	debug (DEBUG_FUNCTION, "Leaving");
	return err;
}

/*
 * Writes the multibyte character at the given buffer onto the terminal
 * Return values:
 * 	0: a valid multibyte sequence has been found and executed
 * 	1: invalid multibyte sequence
 * 	2: uncomplete multibyte character, retry with more data
 * If success, the buffer is increased by the number of bytes processed and
 * the buffer size is decreased accordingly. Else they are left untouched.
 */
int parse_regular_chars (unsigned char **buffer, size_t *buffer_size)
{
	wchar_t current_char;
	int err;
#ifdef MULTIBYTE
	int bytes;
#endif
	
	debug (DEBUG_FUNCTION, "Entering");
#ifdef MULTIBYTE
	bytes = mbtowc (&current_char, (char*)*buffer, *buffer_size);
	switch (bytes) {
		case -1:
			/* invalid or incomplete input */
			/* Note: mbtowc() does not allow to distinguish between
		           the two, hence this dirty hack. mbrtowc() would but
		           is less portable */
			if (*buffer_size < MB_CUR_MAX) err = 2;
			else err = 1;
			debug (DEBUG_FUNCTION, "Leaving");
			return err;
		/*case -1:*/
			/* invalid character encountered */
			/*debug (DEBUG_FUNCTION, "Leaving");
			return 1;
		case -2:*/
			/* incomplete input */
			/*debug (DEBUG_FUNCTION, "Leaving");
			return 2;*/
		default:
			(*buffer_size) -= bytes;
			(*buffer) += bytes;
#ifdef DEBUG
			if (iswprint ((wint_t)current_char)) {
				debug (DEBUG_OUTPUT,
			  	       "Code = 0x%08lX (%09ld)\t\"%lc\"",
			  	       (unsigned long)current_char,
				       (unsigned long)current_char,
				       (wint_t)current_char);
			} else {
				debug (DEBUG_OUTPUT,
				       "Code = 0x%08lX (%09ld)\t(unprintable)",
				       (unsigned long)current_char,
				       (unsigned long)current_char);
			}
#endif /* DEBUG */
	}
#else /* MULTIBYTE */
	current_char = (*buffer)[0];
	(*buffer_size)--;
	(*buffer)++;
#ifdef DEBUG
	if (isprint (current_char)) {
		debug (DEBUG_OUTPUT,
		       "Code = 0x%08lX (%09ld)\t\"%c\"",
		       (unsigned long)current_char,
		       (unsigned long)current_char,
		       (int)current_chare);
	} else {
		debug (DEBUG_OUTPUT,
		       "Code = 0x%08lX (%09ld)\t(unprintable)",
		       (unsigned long)current_char,
		       (unsigned long)current_char);
	}
#endif /* DEBUG */
#endif /* MULTIBYTE */
	text_write_char (current_char);
	debug (DEBUG_FUNCTION, "Leaving");
	return 0;
}
